==> cbra-0.19.1.tar.gz: cbra-0.19.1/cbra/ext/webhook/webhookhandler.py <==
"""Declares :class:`WebhookHandler`."""
import base64
import hmac
import inspect
import logging
from typing import Callable
from typing import List
from typing import Literal
from typing import Optional
from typing import Set

import fastapi
import fastapi.responses
from fastapi.params import Depends

import cbra
from cbra.exceptions import WebhookMessageRejected
from cbra.utils import PositionalArgument


class WebhookHandler(cbra.BaseRequestHandler, cbra.DependantRunner):
    """The base class for all webhook handlers.

    The following methods may use :term:`Injected Parameters`:

    - :meth:`get_event_name()`
    - :meth:`handle()`
    - :meth:`verify_message()`

    """
    __module__: str = 'cbra.ext.webhook'
    logger: logging.Logger = logging.getLogger('uvicorn')

    #: The set of event where this handler is interested in. If
    #: :attr:`allowed_events` is not ``None``, then :meth:`get_event_name()`
    #: must also be implemented.
    allowed_events: Optional[Set[str]] = None

    #: Indicates if the webhook has a dynamic component determined by the
    #: path parameter.
    is_dynamic: bool = False

    #: Specifies the name of the signature header, holding the HMAC of the
    #: message.
    signature_header: str = None

    #: The encoding of the signature. Commonly used encodings are hexadecimal
    #: and URL-encoded base64. Default is ``'hex'``.
    signature_encoding: str = 'hex'

    #: The algorithm used to verify the message HMAC.
    signature_algorithm: Literal['sha256', 'sha384', 'sha512'] = 'sha256'

    #: The set of ignored content type. Use this to ignore webhook pings
    #: and always return 200.
    ignored_content_types: set = set()

    def __init__(self):
        super().__init__(method='POST')

    def add_to_router(self, app: cbra.Application, *args, **kwargs):
        kwargs.setdefault('tags', ['webhooks'])
        return super().add_to_router(app, *args, **kwargs)

    def get_dependants(self) -> List[Callable]:
        """Return the list of additional dependants that must be solved
        when invoking the callable.
        """
        return [self.verify_message]

    def get_event_name(self) -> str:
        """Return the name of the webhook event."""
        if self.allowed_events is not None:
            raise NotImplementedError

    def get_handler_kwargs(self,
        event: str,
        content_type: str,
        *args,
        **kwargs
    ) -> dict:
        """Return the keyword arguments that are passed to the request
        handler function.
        """
        return {'content_type': content_type, 'event': event}

    async def get_signed_bytes(self, request: fastapi.Request) -> bytes:
        """Return a byte-sequence containing the input data for the
        signature.
        """
        return await request.body()

    async def get_secret(self, request: cbra.Request, **kwargs) -> bytes:
        """Return the secret that is used to verify the message HMAC."""
        raise NotImplementedError

    def get_signature_parameters(self):
        sig = inspect.signature(self.handle)
        params = [
            PositionalArgument(
                name='event',
                annotation=str,
                default=Depends(self.get_event_name)
            ),
            PositionalArgument(
                name='content_type',
                annotation=Optional[str],
                default=cbra.Header(None, alias='Content-Type')
            )
        ]
        if self.signature_header:
            params.append(
                PositionalArgument(
                    name='signature',
                    annotation=str,
                    default=cbra.Header(None, alias=self.signature_header)
                )
            )
        if self.is_dynamic:
            params.append(PositionalArgument('webhook_id', annotation=str))
        params.extend([
            x for x in sig.parameters.values()
            if x.name not in self._ignored_parameters
            and not isinstance(x.default, Depends)
        ])
        return params

    async def on_dropped(self,
        request: fastapi.Request,
        event: Optional[str]
    ):
        """Invoked when an incoming message is dropped. This method may raise
        :class:`cbra.exceptions.CanonicalException` to return an error message
        to the client, or return a new :class:`cbra.Response` object.

        If the return value is ``None``, then the default response code is
        used.
        """
        pass

    def parse_signature(self, value: Optional[str]) -> bytes:
        """Parses the header specified by :attr:`signature_header` and decodes
        the content using :attr:`signature_encoding`.
        """
        if self.signature_encoding == 'hex':
            f = bytes.fromhex
        elif self.signature_encoding == 'base64':
            f = base64.b64decode
        else:
            raise NotImplementedError(
                f"Unknown encoding {self.signature_encoding}"
            )
        try:
            return f(value)
        except TypeError:
            self.logger.error("Missing %s header", self.signature_header)
            raise
        except ValueError:
            self.logger.error("Could not decode header %s as %s",
                self.signature_header, self.signature_encoding)
            raise

    async def run_handle(self,
        request: fastapi.Request,
        response: fastapi.Response,
        event: str,
        content_type: str,
        *args, **kwargs
    ):
        """Perform some pre-handle checks and provide to invoke the
        :meth:`handle()` method.
        """
        response = None
        if content_type in self.ignored_content_types:
            response = await self.on_dropped(request, None)
        elif not self.wants_event(request, event):
            response = await self.on_dropped(request, event)
        else:
            await self.run_dependant(
                request=request,
                handle=self.handle,
                body=await request.json(),
                path=request.url.path
            )
        return response or cbra.Response(status_code=200)

    async def verify_message(self,
        request: cbra.Request,
        webhook_id: int = None
    ):
        """Hook to implement message verification. This method is expected
        for raise a fatal exception if verification fails.
        """
        if self.signature_header is not None:
            try:
                signature = self.parse_signature(
                    request.headers.get(self.signature_header)
                )
                digest = hmac.digest(
                    key=await self.get_secret(request),
                    msg=await self.get_signed_bytes(request),
                    digest=self.signature_algorithm
                )
                if not hmac.compare_digest(signature, digest):
                    self.logger.error("The signature did not validate.")
                    raise WebhookMessageRejected
            except (TypeError, ValueError):
                raise WebhookMessageRejected

    def wants_event(self, request: fastapi.Request, event: str) -> bool:
        """Return a boolean indicating if the handler is interested in the
        event.
        """
        return self.allowed_events is None or (event in self.allowed_events)

==> cbra-0.50.0.tar.gz: cbra-0.50.0/cbra/ext/webhook/webhookhandler.py <==
"""Declares :class:`WebhookHandler`."""
import base64
import hmac
import inspect
import logging
from typing import Callable
from typing import List
from typing import Literal
from typing import Optional
from typing import Set

import fastapi
import fastapi.responses
from fastapi.params import Depends

import cbra
from cbra.exceptions import Forbidden
from cbra.exceptions import WebhookMessageRejected
from cbra.utils import PositionalArgument


class WebhookHandler(cbra.handler.BaseRequestHandler, cbra.DependantRunner):
    """The base class for all webhook handlers.

    The following methods may use :term:`Injected Parameters`:

    - :meth:`get_event_name()`
    - :meth:`handle()`
    - :meth:`verify_message()`

    """
    __module__: str = 'cbra.ext.webhook'
    logger: logging.Logger = logging.getLogger('uvicorn')

    #: The set of event where this handler is interested in. If
    #: :attr:`allowed_events` is not ``None``, then :meth:`get_event_name()`
    #: must also be implemented.
    allowed_events: Optional[Set[str]] = None

    #: Indicates if the webhook has a dynamic component determined by the
    #: path parameter.
    is_dynamic: bool = False

    #: Specifies the name of the signature header, holding the HMAC of the
    #: message.
    signature_header: str = None

    #: The encoding of the signature. Commonly used encodings are hexadecimal
    #: and URL-encoded base64. Default is ``'hex'``.
    signature_encoding: str = 'hex'

    #: The algorithm used to verify the message HMAC.
    signature_algorithm: Literal['sha256', 'sha384', 'sha512'] = 'sha256'

    #: The set of ignored content type. Use this to ignore webhook pings
    #: and always return 200.
    ignored_content_types: set = set()

    def __init__(self):
        super().__init__(method='POST')

    def add_to_router(self, app: cbra.Application, *args, **kwargs):
        kwargs.setdefault('tags', ['webhooks'])
        return super().add_to_router(app, *args, **kwargs)

    def get_dependants(self) -> List[Callable]:
        """Return the list of additional dependants that must be solved
        when invoking the callable.
        """
        return [self.verify_message]

    def get_event_name(self, request: fastapi.Request) -> str:
        """Return the name of the webhook event."""
        if self.allowed_events is not None:
            raise NotImplementedError

    def get_handler_kwargs(self,
        event: str,
        content_type: str,
        *args,
        **kwargs
    ) -> dict:
        """Return the keyword arguments that are passed to the request
        handler function.
        """
        return {'content_type': content_type, 'event': event}

    async def get_signed_bytes(self, request: fastapi.Request) -> bytes:
        """Return a byte-sequence containing the input data for the
        signature.
        """
        return await request.body()

    async def get_secret(self, request: cbra.Request, **kwargs) -> bytes:
        """Return the secret that is used to verify the message HMAC."""
        raise NotImplementedError

    def get_signature_parameters(self):
        sig = inspect.signature(self.handle)
        params = [
            PositionalArgument(
                name='event',
                annotation=str,
                default=fastapi.Depends(self.get_event_name)
            ),
            PositionalArgument(
                name='content_type',
                annotation=Optional[str],
                default=cbra.Header(None, alias='Content-Type')
            )
        ]
        if self.signature_header:
            params.append(
                PositionalArgument(
                    name='signature',
                    annotation=str,
                    default=cbra.Header(None, alias=self.signature_header)
                )
            )
        if self.is_dynamic:
            params.append(
                PositionalArgument(
                    'webhook_id',
                    annotation=str,
                    default=cbra.Path(...)
                )
            )
        params.extend([
            x for x in sig.parameters.values()
            if x.name not in self._ignored_parameters
            and not isinstance(x.default, Depends)
            and not (self.is_dynamic and x.name == 'webhook_id')
        ])
        return params

    async def handle(self):
        """Handle the incoming webhook message."""
        raise NotImplementedError

    async def on_dropped(self,
        request: fastapi.Request,
        event: Optional[str]
    ):
        """Invoked when an incoming message is dropped. This method may raise
        :class:`cbra.exceptions.CanonicalException` to return an error message
        to the client, or return a new :class:`cbra.Response` object.

        If the return value is ``None``, then the default response code is
        used.
        """
        pass # pragma: no cover

    def parse_signature(self, value: Optional[str]) -> bytes:
        """Parses the header specified by :attr:`signature_header` and decodes
        the content using :attr:`signature_encoding`.
        """
        if self.signature_encoding == 'hex':
            f = bytes.fromhex
        elif self.signature_encoding == 'base64':
            f = base64.b64decode
        else:
            raise NotImplementedError(
                f"Unknown encoding {self.signature_encoding}"
            )
        try:
            return f(value)
        except TypeError:
            self.logger.error("Missing %s header", self.signature_header)
            raise
        except ValueError:
            self.logger.error("Could not decode header %s as %s",
                self.signature_header, self.signature_encoding)
            raise

    async def run_handle(self,
        request: fastapi.Request,
        response: fastapi.Response,
        event: str,
        content_type: str,
        *args, **kwargs
    ):
        """Perform some pre-handle checks and provide to invoke the
        :meth:`handle()` method.
        """
        response = None
        if content_type in self.ignored_content_types: # pragma: no cover
            response = await self.on_dropped(request, None)
        elif not self.wants_event(request, event):
            raise Forbidden
        else:
            await self.run_dependant(
                request=request,
                handle=self.handle,
                body=await request.json(),
                path=self.path
            )
        return response or cbra.Response(status_code=200)

    async def verify_message(self,
        request: cbra.Request,
        webhook_id: int = None
    ):
        """Hook to implement message verification. This method is expected
        for raise a fatal exception if verification fails.
        """
        if self.signature_header is not None:
            try:
                signature = self.parse_signature(
                    request.headers.get(self.signature_header)
                )
                digest = hmac.digest(
                    key=await self.get_secret(request),
                    msg=await self.get_signed_bytes(request),
                    digest=self.signature_algorithm
                )
                if not hmac.compare_digest(signature, digest):
                    self.logger.error("The signature did not validate.")
                    raise WebhookMessageRejected
            except (TypeError, ValueError):
                raise WebhookMessageRejected

    def wants_event(self, request: fastapi.Request, event: str) -> bool:
        """Return a boolean indicating if the handler is interested in the
        event.
        """
        return self.allowed_events is None or (event in self.allowed_events)
